# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Facility for removing old or QAFailed files and directories from the Workspaces lustre areas
* Download area clean up is handled by the AAT/PPI mr_clean system
"""
import abc
import argparse
import datetime
import logging
import os
import shutil
import sys
from argparse import ArgumentParser, Namespace
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Set, TypedDict

import requests
from dateutil.relativedelta import relativedelta
from pycapo import CapoConfig
from ws_annihilator.connections import MDDBConnector, send_notification

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))
FAILURE_NOTIFICATION_TEMPLATE_NAME = "ws_annihilator_failure"


class Areas(Enum):
    CACHE = "cache"
    SPOOL = "spool"
    STAGING = "staging"
    STORAGE = "storage"


class TypedAnnihilators(Enum):
    """Sometimes we require cleanup times to be different by workflow type"""

    CURATOR = "curator"
    CAL = "std_calibration"
    CMS = "std_cms_imaging"
    RESTORE_IMG = "std_restore_imaging"


Settings = TypedDict(
    "Settings",
    notification_url=str,
    failure_notification_email=str,
    workflow_url=str,
    cache=str,
    spool=str,
    storage=str,
    staging=str,
    cache_days=str,
    spool_days=str,
    staging_days=str,
    storage_days=str,
    calibration_days=str,
    image_days=str,
)


def _get_settings(profile: str) -> Settings:
    """
    Get annihilator capo settings

    :return:
    """
    config = CapoConfig(profile=profile)

    annihilator_settings = config.settings("edu.nrao.workspaces.AnnihilatorSettings")
    ingestion_settings = config.settings("edu.nrao.workspaces.IngestionSettings")
    notification_settings = config.settings("edu.nrao.workspaces.NotificationSettings")

    workflow_url = config.settings("edu.nrao.workspaces.WorkflowSettings").serviceUrl
    notification_url = notification_settings.serviceUrl
    failure_notification_email = notification_settings.analystEmail
    cache_directory = config.settings("edu.nrao.workspaces.DeliverySettings").cacheWeblogDirectory
    spool_directory = config.settings("edu.nrao.workspaces.ProcessingSettings").rootDirectory
    staging_directory = ingestion_settings.stagingDirectory
    storage_directory = ingestion_settings.storageDirectory

    return {
        "workflow_url": workflow_url,
        "notification_url": notification_url,
        "failure_notification_email": failure_notification_email,
        "cache": cache_directory,
        "spool": spool_directory,
        "storage": storage_directory,
        "staging": staging_directory,
        "cache_days": annihilator_settings.keepCacheForDays,
        "spool_days": annihilator_settings.keepSpoolForDays,
        "staging_days": annihilator_settings.keepStagingForDays,
        "storage_days": annihilator_settings.keepStorageForDays,
        "calibration_days": annihilator_settings.keepNonSrdpCalSpoolForDays,
        "image_days": annihilator_settings.keepImageSpoolForDays,
    }


def _arg_parser() -> ArgumentParser:
    """
    Create parser for annihilator

    :return:
    """
    arg_parser = ArgumentParser(
        description="Workspaces Directory Annihilator",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    arg_parser.add_argument(
        "-p",
        "--profile",
        action="store",
        required=True,
        help="Specify which environment to run on. Required. Example: dsoc-dev",
    )
    arg_parser.add_argument(
        "--dry-run",
        action="store_true",
        dest="dry_run",
        required=False,
        help="Dry run, does not actually delete directories nor set workflow as cleaned",
    )
    arg_parser.add_argument(
        "-a",
        "--all",
        action="store_true",
        required=False,
        help="Run annihilation on workspaces spool, staging, and storage areas.",
    )
    arg_parser.add_argument(
        "-c",
        "--cache",
        action="store_true",
        required=False,
        help="Run annihilation on workspaces cache area.",
    )
    arg_parser.add_argument(
        "--curator",
        action="store_true",
        required=False,
        help="Run annihilation on completed curation directories only.",
    )
    arg_parser.add_argument(
        "--calibration",
        action="store_true",
        required=False,
        help="Run annihilation on spool directories from non-SRDP std_calibration only.",
    )
    arg_parser.add_argument(
        "--image",
        action="store_true",
        required=False,
        help="Run annihilation on spool directories from SRDP std_*_images and their calibrations only.",
    )
    return arg_parser


class Annihilator:
    """
    Class for removing workspaces directories that are no longer in use and past a given date

    - Retrieves a list of stale request directories to annihilate
    - Checks if directories still exist
        (DAs might manually clean up directories until scheduling system is available)
    - If directory exists in the workspaces' area for the specified profile, annihilates directory
        and sets 'cleaned' flag on the associated request
    - If directory doesn't exist, sets 'cleaned' flag on associated request
    """

    def __init__(self, args: Namespace, settings: Settings):
        self._args = args
        self._settings = settings
        self._area_to_clean = None
        self._flag_cleaned = False
        self.clean_url = self._settings.get("workflow_url") + "/workflows/requests/cleaned"

    def determine_path(self) -> str:
        """
        return path to the location from capo settings

        :return: path string
        """
        return self._settings.get(self._area_to_clean.value.lower())

    @staticmethod
    def path_valid(constructed_path: str, base_path: str) -> bool:
        """
        Check if the constructed path is valid for annihilation

        :param constructed_path: path to verify
        :param base_path: base path which directory should be annihilated from
        :return: True if path is valid for annihilation
        """

        return (
            constructed_path != base_path
            and constructed_path != base_path + "/"
            and len(constructed_path) > len(base_path)
        )

    def get_days_to_keep(self, key: str) -> str:
        """
        Return the number of days to keep directories in a specified area from capo settings

        :param key: Key for looking up the capo value from the _get_settings dict.
        :return: Number of days as a string
        """
        return self._settings.get(key.lower() + "_days")

    @abc.abstractmethod
    def get_stale_directories(self) -> List[str]:
        """
        It is possible for processing to take multiple days to several weeks. While a job is running in HTCondor, it's
        parent workspaces directory in the spool area won't update other than the HTCondor log file(s).
        We want to be sure that only stale directories are annihilated.

        It is also possible for the dev and test systems to pick up production directories due to database copying.
        For this reason we only want to use the name of the directory itself and strip off the rest of the path.

        :return: A list of directory names to be removed
        """
        pass

    def get_stale_directories_from_query(self, query: str, params: Dict) -> List[str]:
        """
        Common functionality to query the database for stale directories, heavily assumes results will be directories
        (typically want workflow_requests.results_dir).

        :param query: SQL query to execute.
        :param params: Params to fill in query.
        :return: List of stale directories.
        """
        name_list = []
        conn = None

        try:
            conn = MDDBConnector()
            cur = conn.cursor()
            cur.execute(query, params)

            data = cur.fetchall()
            if data:
                for wf_dir in data:
                    # Ignore empty directory names and those from a different environment
                    completed_wf = "".join(wf_dir)
                    if completed_wf and not completed_wf.isspace() and self._args.profile in completed_wf:
                        name_list.append(Path(completed_wf).name)
        except Exception as e:
            logger.error(f"ERROR: Unable to connect to metadata database. %s", exc_info=e)

        if conn is not None:
            conn.close()

        return name_list

    def annihilate_directory(self, directory: str) -> bool:
        """
        Annihilate specified directory

        :param directory: directory to remove
        :return: True if annihilation succeeded, false otherwise
        """

        path = self.determine_path()
        clean_me = path + "/" + directory

        try:
            if self._args.dry_run:
                logger.info(f"Dry run! Would delete {clean_me}")
            else:
                shutil.rmtree(directory)
        except PermissionError as pe:
            logger.error(
                f"ERROR: Permission Denied on directory {clean_me}. "
                f"Please check directory for non-vlapipe owned additions. %s",
                exc_info=pe,
            )
            return False
        except Exception as e:
            logger.error(f"ERROR: {clean_me} encountered the following problem: %s", exc_info=e)
            return False

        # After annihilation set cleaned flag...
        if self._flag_cleaned and not self._args.dry_run:
            logger.info(f"Annihilation complete for {directory}, setting 'cleaned' flag on workflow request...")
            self.set_cleaned(clean_me)
        else:
            logger.info(f"Annihilation complete for {directory} in {self._area_to_clean}.")
        return True

    def set_cleaned(self, directory: str):
        """
        Set the 'cleaned' flag on the request of the specified directory

        :param directory: the directory that was annihilated
        :return:
        """
        json_payload = {"directory": directory, "update_to": "True"}

        return requests.put(self.clean_url, json=json_payload)

    def annihilate(self) -> Set[str]:
        """
        Remove stale directories from targeted workspaces areas

        :return:
        """
        failed_directories = set()
        stale = self.get_stale_directories()

        # If stale directories were found, annihilate them
        if stale:
            path = self.determine_path()

            # change to area to clean
            os.chdir(path)

            for directory in stale:
                dir_path = path + "/" + directory
                if self.path_valid(dir_path, path):
                    if not Path(dir_path).exists():
                        if self._flag_cleaned and not self._args.dry_run:
                            logger.info(
                                f"directory {dir_path} not found, setting 'cleaned' flag on workflow request..."
                            )
                            self.set_cleaned(dir_path)
                        else:
                            logger.info(f"directory {dir_path} not found in {self._area_to_clean}, skipping...")
                    else:
                        logger.info(f"found directory {directory}, annihilating...")
                        did_succeed = self.annihilate_directory(directory)
                        if did_succeed is False:
                            failed_directories.add(dir_path)
                else:
                    logger.error(f"Error: Received invalid directory '{dir_path}' as input for annihilation. Aborting.")
                    exit(1)
        return failed_directories


class SpoolAnnihilator(Annihilator):
    """
    Annihilates stale spool directories from stale workflow requests.
    """

    def __init__(self, args: Namespace, settings: Settings):
        super().__init__(args, settings)
        self._area_to_clean = Areas.SPOOL
        self._flag_cleaned = True

    @property
    def stale_directory_url(self) -> str:
        return (
            self._settings.get("workflow_url")
            + "/workflows/requests/stale/"
            + self.get_days_to_keep(self._area_to_clean.value)
        )

    def get_stale_directories(self) -> List[str]:
        """
        It is possible for processing to take multiple days to several weeks. While a job is running in HTCondor, it's
        parent workspaces directory in the spool area won't update other than the HTCondor log file(s).
        We want to be sure that only stale directories are annihilated.

        It is also possible for the dev and test systems to pick up production directories due to database copying.
        For this reason we only want to use the name of the directory itself and strip off the rest of the path.

        :return: A list of directory names to be removed
        """

        response = requests.get(self.stale_directory_url)
        dir_list = response.json()["stale_dirs"]

        name_list = []
        # catch empty string response
        if dir_list:
            for stale in dir_list.split(","):
                # catch empty string directory or whitespace only directory or VLASS directory (VLASS will handle VLASS)
                if stale and not stale.isspace() and not stale.startswith("VLASS"):
                    name_list.append(Path(str(stale)).name)

        return name_list


class CacheAnnihilator(SpoolAnnihilator):
    """
    Annihilates stale cache directories from stale workflow requests (uses the same stales as the SpoolAnnihilator).
    """

    def __init__(self, args: Namespace, settings: Settings):
        super().__init__(args, settings)
        self._area_to_clean = Areas.CACHE
        self._flag_cleaned = False


class StorageAnnihilator(SpoolAnnihilator):
    """
    Annihilates stale storage directories from stale workflow requests (uses the same stales as the SpoolAnnihilator).
    """

    def __init__(self, args: Namespace, settings: Settings):
        super().__init__(args, settings)
        self._area_to_clean = Areas.STORAGE
        self._flag_cleaned = False


class StagingAnnihilator(Annihilator):
    """
    Annihilates stale staging directories.
    """

    def __init__(self, args: Namespace, settings: Settings):
        super().__init__(args, settings)
        self._area_to_clean = Areas.STAGING

    def get_stale_directories(self) -> List[str]:
        """
        Staging directories shouldn't be needed after ingestion is complete, so this simply annihilates all directories
        older than self.get_days_to_keep().

        :return: A list of directory names to be removed
        """
        path = Path(self.determine_path())

        name_list = []
        for x in path.iterdir():
            # Check if the modification date is older than the days to keep
            age_days = (datetime.datetime.now() - datetime.datetime.fromtimestamp(x.stat().st_ctime)).days
            if age_days > int(self.get_days_to_keep(self._area_to_clean.value)):
                name_list.append(x.name)

        return name_list


class WorkflowAnnihilator(SpoolAnnihilator):
    """
    Annihilates stale spool directories from stale workflow requests for a single workflow.
    """

    def __init__(self, workflow: str, args: Namespace, settings: Settings):
        super().__init__(args, settings)
        self._workflow = workflow

    def get_stale_directories(self) -> List[str]:
        """
        Sometimes we need to clear out all directories meeting certain requirements before the standard clean date

        :return: a list of directories to clean
        """
        query = f"""
        SELECT results_dir
        FROM workflow_requests
        WHERE workflow_name = %(wf_name)s AND state = 'Complete' AND cleaned = false
        """

        params = {"wf_name": self._workflow}

        return self.get_stale_directories_from_query(query, params)


class CalibrationAnnihilator(WorkflowAnnihilator):
    """
    Annihilates stale spool directories from stale calibration requests.
    """

    def get_stale_directories(self) -> List[str]:
        """
        A stale calibration request is defined as one that hasn't been flagged as SRDP within self.get_days_to_keep().

        :return: a list of directories to clean
        """
        query = f"""
        SELECT DISTINCT wr.results_dir
        FROM workflow_requests wr
            JOIN capability_executions ce ON ce.current_workflow_request_id = wr.workflow_request_id
            JOIN capability_requests cr ON cr.capability_request_id = ce.capability_request_id
            JOIN capability_versions cv ON cv.capability_request_id = ce.capability_request_id
        WHERE cr.capability_name = %(wf_name)s
            AND cr.state = 'Complete'
            AND cv.parameters->'metadata'->>'is_srdp' = 'false'
            AND wr.cleaned = false
            AND cr.updated_at < %(time_to_keep)s
        """

        params = {
            "wf_name": self._workflow,
            "time_to_keep": (datetime.datetime.now() - relativedelta(days=int(self.get_days_to_keep("calibration"))))
            .strftime('%Y-%m-%d %H:%M:%S'),
        }

        return self.get_stale_directories_from_query(query, params)


class ImageAnnihilator(WorkflowAnnihilator):
    """
    Annihilates stale spool directories from completed image requests and their associated calibrations.
    """

    def get_stale_directories(self) -> List[str]:
        """
        A completed image request is defined as one that is in a completed state and ingested after
        self.get_days_to_keep(). This also removes the corresponding std_calibration request used to generate the image.

        :return: a list of directories to clean
        """
        # Fetch the completed imaging requests
        query = f"""
        SELECT DISTINCT wr.results_dir
        FROM workflow_requests wr
            JOIN capability_executions ce ON ce.current_workflow_request_id = wr.workflow_request_id
            JOIN capability_requests cr ON cr.capability_request_id = ce.capability_request_id
        WHERE cr.capability_name = %(wf_name)s
            AND cr.state = 'Complete'
            AND cr.ingested = true
            AND wr.cleaned = false
            AND cr.updated_at < %(time_to_keep)s
        """

        params = {
            "wf_name": self._workflow,
            "time_to_keep": (datetime.datetime.now() - relativedelta(days=int(self.get_days_to_keep("image"))))
            .strftime('%Y-%m-%d %H:%M:%S'),
        }

        stales = self.get_stale_directories_from_query(query, params)

        # Fetch the associated calibrations
        query = f"""
        SELECT DISTINCT wr.results_dir
        FROM workflow_requests wr
            JOIN capability_executions ce ON ce.current_workflow_request_id = wr.workflow_request_id
            JOIN capability_requests cr ON cr.capability_request_id = ce.capability_request_id
            JOIN capability_versions cv ON cv.capability_request_id = ce.capability_request_id
        WHERE cr.capability_name = %(cal_wf_name)s
            AND cr.state = 'Complete'
            AND wr.cleaned = false
            AND cr.updated_at < %(time_to_keep)s
            AND cr.capability_request_id::text IN (
                SELECT cv.parameters ->>'parent_request_id' as parent_cal
                FROM workflow_requests wr
                    JOIN capability_executions ce ON ce.current_workflow_request_id = wr.workflow_request_id
                    JOIN capability_requests cr ON cr.capability_request_id = ce.capability_request_id
                    JOIN capability_versions cv ON cv.capability_request_id = ce.capability_request_id
                        AND cv.version_number = ce.capability_version_number
                WHERE cr.capability_name = %(wf_name)s
                    AND cr.state = 'Complete'
                    AND cr.ingested = true
                    AND wr.cleaned = false
                    AND cr.updated_at < %(time_to_keep)s
                GROUP BY parent_cal
            )
        """

        params = {
            "wf_name": self._workflow,
            "cal_wf_name": TypedAnnihilators.CAL.value,
            "time_to_keep": (datetime.datetime.now() - relativedelta(days=int(self.get_days_to_keep("image"))))
            .strftime('%Y-%m-%d %H:%M:%S'),
        }

        stales.extend(self.get_stale_directories_from_query(query, params))

        return stales


def run_annihilator(args: Namespace, settings: Settings):
    failed_directories: Set[str] = set()
    if args.all:
        logger.info(f"Starting cleaning of Workspaces Staging for profile {args.profile}")
        current_failures = StagingAnnihilator(args, settings).annihilate()
        failed_directories.update(current_failures)

        logger.info(f"Starting cleaning of Workspaces Storage for profile {args.profile}")
        current_failures = StorageAnnihilator(args, settings).annihilate()
        failed_directories.update(current_failures)

        # Annihilating spool last because it flags the workflow as being cleaned
        logger.info(f"Starting cleaning of Workspaces Spool for profile {args.profile}")
        current_failures = SpoolAnnihilator(args, settings).annihilate()
        failed_directories.update(current_failures)

        logger.info(f"Finished cleaning Workspaces Spool, Staging, and Storage areas for profile {args.profile}!")

        # Clean completed workflows sooner than the stale time limit
        logger.info(f"Starting cleaning of non-SRDP std_calibration workflows for profile {args.profile}")
        current_failures = CalibrationAnnihilator(TypedAnnihilators.CAL.value, args, settings).annihilate()
        failed_directories.update(current_failures)

        logger.info(f"Starting cleaning of completed imaging workflows for profile {args.profile}")
        current_failures = ImageAnnihilator(TypedAnnihilators.CMS.value, args, settings).annihilate()
        failed_directories.update(current_failures)

        logger.info(f"Finished cleaning spool areas for completed calibrations and images for profile {args.profile}!")
    elif args.cache:
        logger.info(f"Starting cleaning of Workspaces Cache for profile {args.profile}")
        failed_directories = CacheAnnihilator(args, settings).annihilate()
        logger.info(f"Finished cleaning Workspaces Cache for profile {args.profile}!")
    elif args.curator:
        logger.info(f"Starting cleaning of completed Curator workflows for profile {args.profile}")
        failed_directories = WorkflowAnnihilator(TypedAnnihilators.CURATOR.value, args, settings).annihilate()
        logger.info(f"Finished cleaning completed Curator workflows for profile {args.profile}!")
    elif args.calibration:
        logger.info(f"Starting cleaning of non-SRDP std_calibration workflows for profile {args.profile}")
        failed_directories = CalibrationAnnihilator(TypedAnnihilators.CAL.value, args, settings).annihilate()
        logger.info(f"Finished cleaning completed non-SRDP std_calibration workflows for profile {args.profile}!")
    elif args.image:
        logger.info(f"Starting cleaning of completed imaging workflows for profile {args.profile}")
        failed_directories = ImageAnnihilator(TypedAnnihilators.CMS.value, args, settings).annihilate()
        logger.info(f"Finished cleaning completed imaging workflows for profile {args.profile}!")
    else:
        logger.info(f"Starting cleaning of Workspaces Spool for profile {args.profile}")
        failed_directories = SpoolAnnihilator(args, settings).annihilate()
        logger.info(f"Finished cleaning Workspaces Spool for profile {args.profile}!")

    if failed_directories:
        logger.info(
            f"One or more directories weren't annihilated successfully, sending email notification to {settings['failure_notification_email']}"
        )
        send_notification(
            notification_url=settings["notification_url"],
            template=FAILURE_NOTIFICATION_TEMPLATE_NAME,
            parameters={
                "failed_directories": list(failed_directories),
                "destination_email": settings["failure_notification_email"],
            },
        )


def main():
    """
    run the annihilator based on provided args

    :return:
    """
    args = _arg_parser().parse_args()
    settings = _get_settings(args.profile)
    run_annihilator(args, settings)
